@@ -27,9 +27,10 @@ import net.sf.json.JSONArray;
 import org.apache.commons.io.IOUtils;
 import org.kohsuke.stapler.bind.JavaScriptMethod;
 import org.kohsuke.stapler.lang.Klass;
+import org.kohsuke.stapler.lang.MethodRef;
 
+import javax.annotation.PostConstruct;
 import javax.servlet.ServletException;
-import static javax.servlet.http.HttpServletResponse.SC_NOT_FOUND;
 import java.io.IOException;
 import java.lang.reflect.Field;
 import java.lang.reflect.InvocationTargetException;
@@ -38,6 +39,8 @@ import java.util.ArrayList;
 import java.util.List;
 import java.util.Map;
 
+import static javax.servlet.http.HttpServletResponse.*;
+
 /**
  * Created one instance each for a {@link Klass},
  * that retains some useful cache about a class and its views.
@@ -75,6 +78,12 @@ public class MetaClass extends TearOffSupport {
      */
     public final WebApp webApp;
 
+    /**
+     * If there's a method annotated with @PostConstruct, that {@link MethodRef} object, linked
+     * to the list of the base class.
+     */
+    private volatile SingleLinkedList<MethodRef> postConstructMethods;
+
     /*package*/ MetaClass(WebApp webApp, Klass<?> klass) {
         this.clazz = klass.toJavaClass();
         this.klass = klass;
@@ -395,6 +404,27 @@ public class MetaClass extends TearOffSupport {
         }
     }
 
+    /**
+     * Returns all the methods in the ancestory chain annotated with {@link PostConstruct}
+     * from those defined in the derived type toward those defined in the base type.
+     *
+     * Normally invocation requires visiting the list in the reverse order.
+     * @since 1.220
+     */
+    public SingleLinkedList<MethodRef> getPostConstructMethods() {
+        if (postConstructMethods ==null) {
+            SingleLinkedList<MethodRef> l = baseClass==null ? SingleLinkedList.<MethodRef>empty() : baseClass.getPostConstructMethods();
+
+            for (MethodRef mr : klass.getDeclaredMethods()) {
+                if (mr.hasAnnotation(PostConstruct.class)) {
+                    l = l.grow(mr);
+                }
+            }
+            postConstructMethods = l;
+        }
+        return postConstructMethods;
+    }
+
     private String getProtectedRole(Field f) {
         try {
             LimitedTo a = f.getAnnotation(LimitedTo.class);
@@ -454,4 +484,6 @@ public class MetaClass extends TearOffSupport {
             // ignore.
         }
     }
+
+    private static final Object NONE = "none";
 }
