@@ -39,6 +39,7 @@ import org.apache.commons.fileupload.servlet.ServletFileUpload;
 import org.jvnet.tiger_types.Lister;
 import org.kohsuke.stapler.bind.BoundObjectTable;
 import org.kohsuke.stapler.lang.Klass;
+import org.kohsuke.stapler.lang.MethodRef;
 
 import javax.servlet.RequestDispatcher;
 import javax.servlet.ServletContext;
@@ -58,6 +59,7 @@ import java.text.SimpleDateFormat;
 import java.util.ArrayList;
 import java.util.Arrays;
 import java.util.Calendar;
+import java.util.Collection;
 import java.util.Collections;
 import java.util.Date;
 import java.util.Enumeration;
@@ -66,7 +68,9 @@ import java.util.List;
 import java.util.Map;
 import java.util.Set;
 import java.util.StringTokenizer;
+import java.util.logging.Logger;
 
+import static java.util.logging.Level.WARNING;
 import static javax.servlet.http.HttpServletResponse.*;
 
 /**
@@ -288,6 +292,10 @@ public class RequestImpl extends HttpServletRequestWrapper implements StaplerReq
     }
 
     public BindInterceptor setBindListener(BindInterceptor bindListener) {
+        return setBindInterceptpr(bindListener);
+    }
+
+    public BindInterceptor setBindInterceptpr(BindInterceptor bindListener) {
         BindInterceptor o = this.bindInterceptor;
         this.bindInterceptor = bindListener;
         return o;
@@ -529,7 +537,12 @@ public class RequestImpl extends HttpServletRequestWrapper implements StaplerReq
          */
         public Object convertJSON(Object o) {
             Object r = bindInterceptor.onConvert(genericType, type, o);
-            if (r!= BindInterceptor.DEFAULT)    return r; // taken over by the listener
+            if (r!= BindInterceptor.DEFAULT)    return r; // taken over by the interceptor
+
+            for (BindInterceptor i : getWebApp().bindInterceptors) {
+                r = i.onConvert(genericType, type, o);
+                if (r!= BindInterceptor.DEFAULT)    return r; // taken over by the interceptor
+            }
 
             if(o==null || o instanceof JSONNull) {
                 // this method returns null if the type is not primitive, which works.
@@ -571,28 +584,7 @@ public class RequestImpl extends HttpServletRequestWrapper implements StaplerReq
                             }
                         }
 
-                        if (actualType==JSONObject.class || actualType==JSON.class) return actualType.cast(j);
-
-                        String[] names = new ClassDescriptor(actualType).loadConstructorParamNames();
-
-                        // the actual arguments to invoke the constructor with.
-                        Object[] args = new Object[names.length];
-
-                        // constructor
-                        Constructor c = findConstructor(actualType, names.length);
-                        Class[] types = c.getParameterTypes();
-                        Type[] genTypes = c.getGenericParameterTypes();
-
-                        // convert parameters
-                        for( int i=0; i<names.length; i++ ) {
-                            try {
-                                args[i] = bindJSON(genTypes[i],types[i],j.get(names[i]));
-                            } catch (IllegalArgumentException e) {
-                                throw new IllegalArgumentException("Failed to convert the "+names[i]+" parameter of the constructor "+c,e);
-                            }
-                        }
-
-                        return invokeConstructor(c, args);
+                        return instantiate(actualType, j);
                     } catch (IllegalArgumentException e) {
                         throw new IllegalArgumentException("Failed to instantiate "+type+" from "+j,e);
                     }
@@ -663,6 +655,116 @@ public class RequestImpl extends HttpServletRequestWrapper implements StaplerReq
         }
     }
 
+    /**
+     * Called after the actual type of the binding is figured out.
+     */
+    private Object instantiate(Class actualType, JSONObject j) {
+        Object r = bindInterceptor.instantiate(actualType,j);
+        if (r!=BindInterceptor.DEFAULT) return r;
+        for (BindInterceptor bi : getWebApp().bindInterceptors) {
+            r = bi.instantiate(actualType,j);
+            if (r!=BindInterceptor.DEFAULT) return r;
+        }
+
+        if (actualType==JSONObject.class || actualType==JSON.class) return actualType.cast(j);
+
+        String[] names = new ClassDescriptor(actualType).loadConstructorParamNames();
+
+        // the actual arguments to invoke the constructor with.
+        Object[] args = new Object[names.length];
+
+        // constructor
+        Constructor c = findConstructor(actualType, names.length);
+        Class[] types = c.getParameterTypes();
+        Type[] genTypes = c.getGenericParameterTypes();
+
+        // convert parameters
+        for( int i=0; i<names.length; i++ ) {
+            try {
+                args[i] = bindJSON(genTypes[i],types[i],j.get(names[i]));
+            } catch (IllegalArgumentException e) {
+                throw new IllegalArgumentException("Failed to convert the "+names[i]+" parameter of the constructor "+c,e);
+            }
+        }
+
+        return injectSetters(invokeConstructor(c, args), j, Arrays.asList(names));
+    }
+
+    /**
+     * @param exclusions
+     *      Properties that are already injected through the constructor, thus not subject of the setter injection.
+     */
+    private <T> T injectSetters(T r, JSONObject j, Collection<String> exclusions) {
+        // try to assign rest of the properties
+        OUTER:
+        for (String key : (Set<String>)j.keySet()) {
+            if (!exclusions.contains(key)) {
+                try {
+                    // try field injection first
+                    for (Class c=r.getClass(); c!=null; c=c.getSuperclass()) {
+                        try {
+                            Field f = c.getDeclaredField(key);
+                            if (f.getAnnotation(DataBoundSetter.class)!=null) {
+                                f.setAccessible(true);
+                                f.set(r, bindJSON(f.getGenericType(), f.getType(), j.get(key)));
+                                continue OUTER;
+                            }
+                        } catch (NoSuchFieldException e) {
+                            // recurse into parents
+                        }
+                    }
+
+                    PropertyDescriptor pd = PropertyUtils.getPropertyDescriptor(r, key);
+                    if (pd==null)   continue;
+
+                    Method wm = pd.getWriteMethod();
+                    if (wm==null)   continue;
+                    if (wm.getAnnotation(DataBoundSetter.class)==null) {
+                        LOGGER.warning("Tried to set value to "+key+" but method "+wm+" is missing @DataBoundSetter");
+                        continue;
+                    }
+
+                    Class<?>[] pt = wm.getParameterTypes();
+
+                    if (pt.length!=1) {
+                        LOGGER.log(WARNING, "Setter method "+wm+" is expected to have 1 parameter");
+                        continue;
+                    }
+
+                    // only invoking public methods for security reasons
+                    wm.invoke(r, bindJSON(wm.getGenericParameterTypes()[0], pt[0], j.get(key)));
+                } catch (IllegalAccessException e) {
+                    LOGGER.log(WARNING, "Cannot access property " + key + " of " + r.getClass(), e);
+                } catch (InvocationTargetException e) {
+                    LOGGER.log(WARNING, "Cannot access property " + key + " of " + r.getClass(), e);
+                } catch (NoSuchMethodException e) {
+                    LOGGER.log(WARNING, "Cannot access property " + key + " of " + r.getClass(), e);
+                }
+            }
+        }
+
+        invokePostConstruct(getWebApp().getMetaClass(r).getPostConstructMethods(), r);
+
+        return r;
+    }
+
+    /**
+     * Invoke PostConstruct method from the base class to subtypes.
+     */
+    private void invokePostConstruct(SingleLinkedList<MethodRef> methods, Object r) {
+        if (methods.isEmpty())  return;
+
+        invokePostConstruct(methods.tail,r);
+        try {
+            methods.head.invoke(r);
+        } catch (InvocationTargetException e) {
+            throw new IllegalArgumentException("Unable to post-construct "+r,e);
+        } catch (IllegalAccessException e) {
+            throw (Error)new IllegalAccessError().initCause(e);
+        }
+
+    }
+
     /**
      * Gets the type of the field/property designate by the given name.
      */
@@ -765,7 +867,6 @@ public class RequestImpl extends HttpServletRequestWrapper implements StaplerReq
                         rsp.sendError(SC_BAD_REQUEST,"This page expects a form submission");
                     else
                         rsp.sendError(SC_BAD_REQUEST,"Nothing is submitted");
-                    rsp.getWriter().close();
                     throw new ServletException("This page expects a form submission but had only " + getParameterMap());
                 } catch (IOException e) {
                     throw new ServletException(e);
@@ -787,4 +888,6 @@ public class RequestImpl extends HttpServletRequestWrapper implements StaplerReq
         if(item==null || item.isFormField())    return null;
         return item;
     }
+
+    private static final Logger LOGGER = Logger.getLogger(RequestImpl.class.getName());
 }
