@@ -23,7 +23,6 @@
 
 package org.kohsuke.stapler;
 
-import com.google.common.collect.ImmutableMap;
 import net.sf.json.JSONObject;
 import org.apache.commons.beanutils.ConversionException;
 import org.apache.commons.beanutils.ConvertUtils;
@@ -33,6 +32,7 @@ import org.apache.commons.beanutils.converters.DoubleConverter;
 import org.apache.commons.beanutils.converters.FloatConverter;
 import org.apache.commons.beanutils.converters.IntegerConverter;
 import org.apache.commons.fileupload.FileItem;
+import org.apache.commons.io.IOUtils;
 import org.kohsuke.stapler.bind.BoundObjectTable;
 
 import javax.servlet.RequestDispatcher;
@@ -45,6 +45,7 @@ import javax.servlet.http.HttpServlet;
 import javax.servlet.http.HttpServletRequest;
 import javax.servlet.http.HttpServletResponse;
 import java.io.DataInputStream;
+import java.io.EOFException;
 import java.io.File;
 import java.io.IOException;
 import java.io.InputStream;
@@ -52,6 +53,7 @@ import java.io.OutputStream;
 import java.io.PrintWriter;
 import java.io.UnsupportedEncodingException;
 import java.lang.reflect.InvocationTargetException;
+import java.net.JarURLConnection;
 import java.net.MalformedURLException;
 import java.net.SocketException;
 import java.net.URL;
@@ -62,6 +64,7 @@ import java.text.SimpleDateFormat;
 import java.util.ArrayList;
 import java.util.Arrays;
 import java.util.Collection;
+import java.util.Collections;
 import java.util.Date;
 import java.util.HashMap;
 import java.util.HashSet;
@@ -153,7 +156,7 @@ public class Stapler extends HttpServlet {
                 }
             }
 
-            resourcePaths = ImmutableMap.copyOf(paths);
+            resourcePaths = Collections.unmodifiableMap(paths);
         } catch (MalformedURLException e) {
             resourcePaths = null; // abort
         }
@@ -163,6 +166,10 @@ public class Stapler extends HttpServlet {
         return webApp;
     }
 
+    /*package*/ void setWebApp(WebApp webApp) {
+        this.webApp = webApp;
+    }
+
     protected @Override void service(HttpServletRequest req, HttpServletResponse rsp) throws ServletException, IOException {
         Thread t = Thread.currentThread();
         final String oldName = t.getName();
@@ -257,6 +264,43 @@ public class Stapler extends HttpServlet {
         private void close() throws IOException {
             stream.close();
         }
+
+        /**
+         * Can't just use {@code connection.getLastModified()} because
+         * of a file descriptor leak in {@code JarURLConnection}.
+         * See http://sourceforge.net/p/freemarker/bugs/189/
+         */
+        public long getLastModified() {
+            if (connection instanceof JarURLConnection) {
+                // There is a bug in sun's jar url connection that causes file handle leaks when calling getLastModified()
+                // Since the time stamps of jar file contents can't vary independent from the jar file timestamp, just use
+                // the jar file timestamp
+                URL jarURL = ((JarURLConnection) connection).getJarFileURL();
+                if (jarURL.getProtocol().equals("file")) {
+                    // Return the last modified time of the underlying file - saves some opening and closing
+                    return new File(jarURL.getFile()).lastModified();
+                } else {
+                    // Use the URL mechanism
+                    URLConnection jarConn = null;
+                    try {
+                        jarConn = jarURL.openConnection();
+                        return jarConn.getLastModified();
+                    } catch (IOException e) {
+                        return -1;
+                    } finally {
+                        if (jarConn != null) {
+                            try {
+                                IOUtils.closeQuietly(jarConn.getInputStream());
+                            } catch (IOException e) {
+                                // ignore this error
+                            }
+                        }
+                    }
+                }
+            } else {
+                return connection.getLastModified();
+            }
+        }
     }
 
     /**
@@ -346,10 +390,10 @@ public class Stapler extends HttpServlet {
      * Serves the specified {@link URLConnection} as a static resource.
      */
     boolean serveStaticResource(HttpServletRequest req, StaplerResponse rsp, OpenConnection con, long expiration) throws IOException {
-        if(con==null)   return false;
+        if (con == null) return false;
         try {
-            return serveStaticResource(req,rsp, con.stream,
-                    con.connection.getLastModified(),
+            return serveStaticResource(req, rsp, con.stream,
+                    con.getLastModified(),
                     expiration,
                     con.connection.getContentLength(),
                     con.connection.getURL().toString());
@@ -493,7 +537,7 @@ public class Stapler extends HttpServlet {
 
                         // ritual for responding to a partial GET
                         rsp.setStatus(HttpServletResponse.SC_PARTIAL_CONTENT);
-                        rsp.setHeader("Content-Range",s+"-"+(e-1)+'/'+contentLength); // end is inclusive.
+                        rsp.setHeader("Content-Range","bytes "+s+"-"+(e-1)+'/'+contentLength); // end is inclusive.
 
                         // prepare to send the partial content
                         DataInputStream dis = new DataInputStream(in);
@@ -627,8 +671,7 @@ public class Stapler extends HttpServlet {
         }
 
         // adds this node to ancestor list
-        AncestorImpl a = new AncestorImpl(req.ancestors);
-        a.set(node,req);
+        AncestorImpl a = new AncestorImpl(req, node);
 
         // try overrides
         if (node instanceof StaplerOverridable) {
@@ -757,7 +800,13 @@ public class Stapler extends HttpServlet {
         return false;
     }
 
-    private static boolean isSocketException(Throwable x) { // JENKINS-10524
+    /**
+     * Used to detect exceptions thrown when writing content that seem to be due merely to a closed socket.
+     * @param x an exception that got caught
+     * @return true if this looks like a closed stream, false for some real problem
+     * @since 1.222
+     */
+    public static boolean isSocketException(Throwable x) { // JENKINS-10524
         if (x == null) {
             return false;
         }
@@ -767,6 +816,16 @@ public class Stapler extends HttpServlet {
         if (String.valueOf(x.getMessage()).equals("Broken pipe")) { // TBD I18N
             return true;
         }
+        // JENKINS-20074: various things that Jetty and other components could throw
+        if (x instanceof EOFException) {
+            return true;
+        }
+        if (x instanceof IOException && "Closed".equals(x.getMessage())) { // org.eclipse.jetty.server.HttpOutput.print
+            return true;
+        }
+        if (x instanceof IOException && "finished".equals(x.getMessage())) { //com.jcraft.jzlib.DeflaterOutputStream.write
+            return true;
+        }
         return isSocketException(x.getCause());
     }
 
